# include <assert.h>
# include "gcc-externs.h"
# include "sdg.h"
# include "bb.h"
# include "ud.h"
# include "regset.ic"

	/* For the 'bb' being processed, keeps track of the ids of defns
	 * within the basic block */
extern int defnId;

	/* defined in 'ud.c' */
extern int numDefns;

	/* defined in '../gcc-2.7.1/flow.c' */
extern int max_regno;

	/* If this is set, then we are currently processing a parallel rtx */
static int inParallelRtx;

	/* The basic block currently being processed */
BasicBlock *currBB;

	/* The current sdg node being processed */
SdgNode *currSdgNode;

	/* List of instructions in this function which are already in
	 * the ld-st subsystem and from which slicing starts */
extern List sliceNodeList;

	/* This is set if an integer return register is used in any instruction */
static int intRetRegUsed;

	/* This is set if an integer argument register is used in any instruction */
static int intArgRegUsed;

	/* This is set if a hard integer register is used in any instruction */
static int hardIntRegUsed;

	/* This is set if a CALL rtx is seen.  Used in marking the
	 * rest of the rtl expressions part of this instruction */
static int insideACall;

	/* If current instr. is a conversion op, which way is it? */
static int intToFpConv;

# define IsAnIntegerHardReg(regNum) \
    (  (regNum) < FIRST_PSEUDO_REGISTER  \
	&& REGNO_REG_CLASS(regNum) == GENERAL_REGS)

# define IsAnIntegerArgumentReg(regNum) \
    ((regNum) >= GP_ARG_FIRST && (regNum) <= GP_ARG_LAST)

# define IsAnIntegerReturnReg(regNum) \
	((regNum) == GP_RETURN)

# ifdef COPYOPT_SLICING

SdgNode *dummyArgNodes[GP_ARG_LAST - GP_ARG_FIRST + 1];

SdgNode *
GetDummyArgumentNode(int regNum)
{
	if (regNum == 0)
		return (SdgNode *)NULL;

	if (  (regNum >= GP_ARG_FIRST && regNum <= GP_ARG_LAST)
	    || (regNum >= FP_ARG_FIRST && regNum <= FP_ARG_LAST))
	{
		return (regNum >= FP_ARG_FIRST) 
					? (SdgNode *)NULL
					: dummyArgNodes[regNum - GP_ARG_FIRST];
	}
	else {
		fprintf(stderr, "Possible USE BEFORE DEF of pseudo %d\n", regNum);
		return (SdgNode *)NULL;
	}
}

void
InitDummyArgumentNodes()
{
	int i;

	for (i = 0; i < GP_ARG_LAST - GP_ARG_FIRST + 1; i++) {
		dummyArgNodes[i] = AllocNewSDGnode((rtx)NULL);
		dummyArgNodes[i]->inLdSt = 1;
		dummyArgNodes[i]->dstUsesIntArgReg = 1;
		dummyArgNodes[i]->dstUsesHardIntReg = 1;
	}
}

void
FreeDummyArgumentNodes()
{
	int i;

	for (i = 0; i < GP_ARG_LAST - GP_ARG_FIRST + 1; i++) {
		FreeList(&dummyArgNodes[i]->srcs);
		FreeList(&dummyArgNodes[i]->dsts);
		free(dummyArgNodes[i]);
	}
}

# endif

void
IdentifyDependenciesInExpr(rtx expr)
{
	Preg     *reg;
	RTX_CODE  code;
	int       regNum;
	rtx       tmp;
	
	switch (code = GET_CODE(expr)) {
			/* The following cases are constant expressions
			 * which do not use any pseudo regs. */
			/* CHECK : Verify that the above is true for 'CONST_DOUBLE' */
	case CONST_INT :
	case CONST_STRING :
	case HIGH :
	case CONST_DOUBLE :
	case LABEL_REF :
		break;

	case SYMBOL_REF :
			/* This could reference the frame/stack pointers
			 * and hence should be in the ld-st subsystem
			 * But, this is fine if the reference is as part
			 * of a load/store address */
		if (!currSdgNode->isMemOp && !insideACall)
			hardIntRegUsed = 1;
		break;

			/* Follow this to check if any SYMBOL_REF's are used */
	case CONST :
		IdentifyDependenciesInExpr(XEXP(expr, 0));
		break;

			/* For these operand types, there is no 
			 * use of a pseudo reg => no dependency edge added. */
	case SCRATCH :
	case CC0 :
	case PC :
		break;

	case MEM :
		if (!insideACall) {
				/* Corresponds to a load/store instruction *
				 * Put this node in the 'sliceNodeList' */
			Append(&sliceNodeList, currSdgNode);
			currSdgNode->isMemOp = 1;
				/* For a store op, the calling function
				 * will reset this flag */
			currSdgNode->isLoadOp = 1;
		}
		IdentifyDependenciesInExpr(XEXP(expr, 0));
		break;

	case REG :
			/* This is a use of the reg 'expr'.  Get its corr. preg
			 * Then, for all reaching defns. of this 'reg',
			 * add edges from each of those defns. to 'currSdgNode' */
		regNum = REGNO(expr);
		if (regNum != STACK_POINTER_REGNUM && regNum != FRAME_POINTER_REGNUM) {
			reg = GetPreg(currBB, regNum);
				/* This could be null for function arguments because
				 * there might be no prior definition of this register
				 * before it is used */
			if (reg != NULL)
				AddSdgEdges(reg, currSdgNode);
# ifdef COPYOPT_SLICING
			else {
				SdgNode *argNode = GetDummyArgumentNode(regNum);
				if (argNode != NULL)
					AddSdgEdge(argNode, currSdgNode);
			}
# endif
		}
			/* Any instruction which uses an integer hard register
			 * has to be in the load-store subsystem */
		if (IsAnIntegerHardReg(regNum)) {
			if (IsAnIntegerReturnReg(regNum))
				intRetRegUsed = 1;
			else if (IsAnIntegerArgumentReg(regNum))
				intArgRegUsed = 1;
			else
				hardIntRegUsed = 1;
		}
# ifdef NO_MULT_DIVS
		if (!currSdgNode->inLdSt && MD_REG_P(regNum)) {
			currSdgNode->inLdSt = 1;
			Append(&sliceNodeList, currSdgNode);
		}
# endif
		break;

	case SUBREG :
		regNum = REGNO(SUBREG_REG(expr));
		if (regNum != STACK_POINTER_REGNUM && regNum != FRAME_POINTER_REGNUM) {
			reg = GetPreg(currBB, regNum);
				/* This could be null for function arguments because
				 * there might be no prior definition of this register
				 * before it is used */
			if (reg != NULL)
				AddSdgEdges(reg, currSdgNode);
# ifdef COPYOPT_SLICING
			else {
				SdgNode *argNode = GetDummyArgumentNode(regNum);
				if (argNode != NULL)
					AddSdgEdge(argNode, currSdgNode);
			}
# endif
		}
			/* Any instruction which uses an integer hard register
			 * has to be in the load-store subsystem */
		if (IsAnIntegerHardReg(regNum)) {
			if (IsAnIntegerReturnReg(regNum))
				intRetRegUsed = 1;
			else if (IsAnIntegerArgumentReg(regNum))
				intArgRegUsed = 1;
			else
				hardIntRegUsed = 1;
		}
		break;

	case STRICT_LOW_PART :
			/* This code can appear only in a SET_DEST.
			 * However, this function is not called for SET_DEST. */
		assert(0);
		break;

			/* Binary arithmetic operations */
	case MULT :
	case DIV : 
	case UDIV :
	case MOD : 
	case UMOD :
		IdentifyDependenciesInExpr(XEXP(expr, 0));
		IdentifyDependenciesInExpr(XEXP(expr, 1));
# ifdef NO_MULT_DIVS
		if (!currSdgNode->inLdSt) {
			currSdgNode->inLdSt = 1;
			Append(&sliceNodeList, currSdgNode);
		}
# endif
		break;
	
			/* Binary arithmetic operations */
	case PLUS :
	case LO_SUM :
	case MINUS :
	case COMPARE :
	case SMIN :
	case SMAX :
	case UMIN :
	case UMAX :
	case AND :
			/* Binary bit operations */
	case IOR :
	case XOR :
	case ASHIFT :
	case LSHIFTRT :
	case ASHIFTRT :
	case ROTATE :
	case ROTATERT :
			/* Binary comparison operations */
	case EQ :
	case NE :
	case GT :
	case GTU :
	case LT :
	case LTU :
	case GE :
	case GEU :
	case LE :
	case LEU :
		IdentifyDependenciesInExpr(XEXP(expr, 0));
		IdentifyDependenciesInExpr(XEXP(expr, 1));
		break;

			/* Unary arithmetic operations */
	case NOT :
	case NEG :
	case ABS :
	case SQRT :
	case FFS :
		IdentifyDependenciesInExpr(XEXP(expr, 0));
		break;

			/* Call.  Indirect calls use a register */
	case CALL :
		insideACall = 1;
		IdentifyDependenciesInExpr(XEXP(expr, 0));
		break;

			/* Branches(conditional jumps) 
			 * condition   => comparison expression
			 * then-clause => most probably a 'label_ref'
			 * else-clause => most probably 'pc'
			 */
	case IF_THEN_ELSE :
		IdentifyDependenciesInExpr(XEXP(expr, 0));	/* condition */
		IdentifyDependenciesInExpr(XEXP(expr, 1));	/* then-clause */
		IdentifyDependenciesInExpr(XEXP(expr, 2));	/* else-clause */
		break;

			/* "More general if-then-else", so says the rtl info node
			 * Might be used to support a switch.
			 * It also says that this is not valid as an instruction 
			 * pattern currently.  Only supported for describing 
			 * instruction attributes in the m/c description. */
	case COND :
		fprintf(stderr, 
				"COND is not supposed to appear as an instr. pattern\n");
		abort();
		break;

			/* Bit-field operations */
			/* The bits are extracted from XEXP(expr, 0) */
	case SIGN_EXTRACT :
	case ZERO_EXTRACT :
		IdentifyDependenciesInExpr(XEXP(expr, 0)); 
		break;

			/* Conversions within the same mode - only precision changes */
	case SIGN_EXTEND :
	case ZERO_EXTEND :
	case FLOAT_EXTEND :
	case FLOAT_TRUNCATE :
	case TRUNCATE :
		IdentifyDependenciesInExpr(XEXP(expr, 0)); 
		break;

			/* Conversion between integer and floating point modes
             * This means that there is communication between the two
			 * register files thru' this instruction.
			 * => For our purposes, this is a copy node, though the
			 * instruction first produces a value and then copies it */
	case FIX :
	case UNSIGNED_FIX :
		tmp = XEXP(expr, 0);
		IdentifyDependenciesInExpr(tmp);
		assert(INTEGRAL_MODE_P(GET_MODE(expr)) && FLOAT_MODE_P(GET_MODE(tmp)));
		currSdgNode->isAConvNode = 1;
		intToFpConv = 0;
		break;

	case FLOAT :
	case UNSIGNED_FLOAT :
		tmp = XEXP(expr, 0);
		IdentifyDependenciesInExpr(tmp);
		assert(INTEGRAL_MODE_P(GET_MODE(tmp)) && FLOAT_MODE_P(GET_MODE(expr)));
		currSdgNode->isAConvNode = 1;
		intToFpConv = 1;
		break;

			/* Unhandled codes */
	default :
		fprintf(stderr, "Unhandled case in \'IdentifyDependenciesInExpr\'\n");
		abort();
	}
}

void
CheckIfDestUsesIntHardReg(rtx reg)
{
/* Any instruction which uses an integer hard register
 * has to be in the load-store subsystem */

	int regNum = REGNO(reg);

	if (IsAnIntegerHardReg(regNum)) {
			/* Note down that this node uses an argument/return reg for 
			 * its dest.  Later, if it is found out that this node doesn't
			 * receive values from other nodes, this node is put in the
			 * ld-st subsystem saving a 'copy' */
		if (IsAnIntegerArgumentReg(regNum))
			currSdgNode->dstUsesIntArgReg = 1;
		else if (IsAnIntegerReturnReg(regNum))
			currSdgNode->dstUsesIntRetReg = 1;
		else {
			currSdgNode->dstUsesHardIntReg = 1;
			if (!hardIntRegUsed) {
					/* If 'hardIntRegUsed' is set, then this node
					 * will already have been put into the list */
				currSdgNode->inLdSt = 1;
				Append(&sliceNodeList, currSdgNode);
			}
				/* However, if the mode the hard int reg is used is
				 * a fp mode, then this is treated as a conv node */
			if (FLOAT_MODE_P(GET_MODE(reg))) {
				fprintf(stderr, " FOUND ONE!! node id - %d\n", currSdgNode->id);
					/* BE CONSERVATIVE.  
					 * Dont know which way conversion this is
					 * So, force this into LdSt by not marking it
					 * a conv node even though it could be one. */
			}
		}
	}
}

void
IdentifyDependenciesIn_SET_rtx(rtx insn)
{
/* Firstly, identify dependendencies in the srcs/dests.
 * Next, update 'reaching defn' information, For 'PARELLEL' rtx, 
 * this info. is not updated until all exprs. in the parallel 
 * vector are processed. */

	rtx dest;
	int regNum;

		/* Initialise */
	hardIntRegUsed = 0;
	intRetRegUsed = 0;
	intArgRegUsed = 0;
	insideACall = 0;
	intToFpConv = 0;

		/* Takes care of all 'SET' instructions 
		 * loads, stores, arithmetic ops, even indirect jumps
		 * The source should be processed first.
		 * Later code depends on this */
	IdentifyDependenciesInExpr(SET_SRC(insn));
	if (hardIntRegUsed) {
		currSdgNode->srcUsesHardIntReg = 1;
		currSdgNode->inLdSt = 1;

			/* If the integer hard register is being used
			 * in the FP mode, then this insn is treated as a conv op */
		if (!currSdgNode->isMemOp && FLOAT_MODE_P(GET_MODE(SET_SRC(insn)))) {
			fprintf(stderr, " FOUND ONE!! node id - %d\n", currSdgNode->id);
			assert(!currSdgNode->isAConvNode || intToFpConv);
			currSdgNode->isAConvNode = 1;
			intToFpConv = 1;
		}

			/* If the insn. was a mem op, it will have been put
			 * in the list already */
		if (!currSdgNode->isMemOp)
			Append(&sliceNodeList, currSdgNode);
	}
	else if (intArgRegUsed)
		currSdgNode->srcUsesIntArgReg = 1;
	else if (intRetRegUsed)
		currSdgNode->srcUsesIntRetReg = 1;


	if (currSdgNode->isAConvNode) {
			/* This node has been split into 2.  Move on to the
			 * node that is the defn node */
		if (intToFpConv) {
			currSdgNode->isAConvNode = 0;
			currSdgNode->next->isAConvNode = 1;
		}
		currSdgNode = currSdgNode->next;
	}

	assert(   !currSdgNode->isLoadOp
	       || (currSdgNode->isLoadOp & currSdgNode->isAddr));

		/* If this node was an address node(when this is a load, for ex.)
		 * move to the next node */
	if (currSdgNode->isAddr) {
		currSdgNode = currSdgNode->next;
		currSdgNode->isLoadOp = 1;
		currSdgNode->isMemOp = 1;
	}

	dest = SET_DEST(insn);
	insideACall = 0;
	switch (GET_CODE(dest)) {
		case MEM :
			if (!IsEmpty(currSdgNode->srcs)) {
					/* Split the store instruction into 2 nodes
					 * The store address and the store value.
					 * This eliminates all special cases in the slicing code */
				SdgNode *stAddr = AllocNewSDGnode(currSdgNode->instr);

				stAddr->inCompoundRtx = currSdgNode->inCompoundRtx;
				stAddr->compoundRtx   = currSdgNode->compoundRtx;
				stAddr->nodeWeight    = currSdgNode->nodeWeight;

				stAddr->isAddr = 1;
				currSdgNode->isStoreVal = 1;

					/* Add this new node into the graph */
				stAddr->next      = currSdgNode->next;
				currSdgNode->next = stAddr;
				if (sdg.tail == currSdgNode)
					sdg.tail = stAddr;

					/* Reset currSdgNode */
				stAddr->inLdSt = 1;
				currSdgNode = stAddr;
			}

				/* Corresponds to store address computation */
			IdentifyDependenciesInExpr(dest);
			currSdgNode->isLoadOp = 0;
			currSdgNode->isStoreOp = 1;
			break;

		case REG :
			regNum = REGNO(dest);
			CheckIfDestUsesIntHardReg(dest);
			if (!inParallelRtx)
				UpdateReachingDefnInCurrBB(regNum);
			break;

		case SUBREG :
			regNum = REGNO(SUBREG_REG(dest));
			CheckIfDestUsesIntHardReg(SUBREG_REG(dest));
			if (!inParallelRtx)
				UpdateReachingDefnInCurrBB(regNum);
			break;

		case STRICT_LOW_PART :
			regNum = REGNO(SUBREG_REG(XEXP(dest, 0)));
			CheckIfDestUsesIntHardReg(SUBREG_REG(XEXP(dest, 0)));
			if (!inParallelRtx)
				UpdateReachingDefnInCurrBB(regNum);
			break;

		default :
			break;
	}
}

void
IdentifyDependenciesIn_UNSPEC_rtx(rtx insn)
{
/* UNSPEC/UNSPEC_VOLATILE have operands which're used in some
 * m/c specific operation(s).  These are considered to define/use
 * all their operands */

	int   i, regNum;
	rtx   elt;
	Preg *reg;

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		elt = XVECEXP(insn, 0, i);
		if (REG_P(elt)) {
				/* This is a use of 'elt'.  
				 * Get the preg corr. to 'elt'.
				 * Then, for all reaching defns. of this 'reg',
				 * add edges from each of those defns. to 'currSdgNode' */
			regNum = REGNO(elt);
			if (   regNum != STACK_POINTER_REGNUM 
				&& regNum != FRAME_POINTER_REGNUM)
			{
				reg = GetPreg(currBB, regNum);
					/* This could be null for function arguments because
					 * there might be no prior definition of this register
					 * before it is used */
				if (reg != NULL) {
					AddSdgEdges(reg, currSdgNode);
						/* Now, update 'reg->tmp' to reflect this 
						 * new definition */
					Defn_ClearSet(reg->tmp);
					Defn_SetBit(reg->tmp, defnId);
					defnId++;
				}
			}
		}
	}
}

void
IdentifyDependenciesIn_SEQUENCE_rtx(rtx insn)
{
/* SEQUENCE holds a vector of instructions */

	int 	 i;
	rtx      instr;
	RTX_CODE code;
	extern void IdentifyDependenciesInInstr(rtx);

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		instr = XVECEXP(insn, 0, i);
		code = GET_CODE(instr);
		if (code == INSN || code == CALL_INSN || code == JUMP_INSN) {
			IdentifyDependenciesInInstr(instr);
			assert(instr == currSdgNode->instr); /* Sanity check */
			currSdgNode = currSdgNode->next;
		}
	}
}

void
IdentifyDependenciesIn_PARALLEL_rtx(rtx insn)
{
/* PARALLEL holds a vector of expressions each
 * of which is a side-effect expression */

	int   i;
	rtx   elt;
	Preg *reg;

	extern void Handle_PARALLEL_rtx(rtx insn);

	inParallelRtx = 1;

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		elt = XVECEXP(insn, 0, i);

		assert(elt == currSdgNode->instr); /* Sanity check */

		switch (GET_CODE(elt)) {
		case SET :
			IdentifyDependenciesIn_SET_rtx(elt);
			break;

		case CLOBBER :
		case RETURN :
		case USE :
		case CALL :
				/* None of these 'rtl's use any register
				 * 'USE' is not considered since it will be
				 * deleted in later phases */
			break;

				/* Surprise! Surprise!  This seems to be generated
				 * during optimizations */
		case CONST_INT :
			break;

		default :
			fprintf(stderr, "Unhandled case in \'IdentifyDependenciesIn_PARALLEL_rtx\'\n");
			abort();
		}

		currSdgNode = currSdgNode->next;
	}

	inParallelRtx = 0;

		/* Now, update reaching defn. info for this instr. */
	Handle_PARALLEL_rtx(insn);
}

void
IdentifyDependenciesInInstr(rtx insn)
{
/* Now, we are trying to identify both uses & defns.
 * This will help us find register dependencies */

	rtx   body, dest;

	switch (GET_CODE(insn)) {
	case ADDR_VEC :
	case ADDR_DIFF_VEC :
	case RETURN :
	case USE :
				/* USE is not needed because it doesn't represent an
				 * actual instruction.  It will be deleted during later
				 * phases */
		break;

			/* Indirect calls could use a register */
	case CALL :
		insideACall = 1;
		IdentifyDependenciesInExpr(XEXP(insn, 0));
		break;

	case CLOBBER :
			/* Update 'defn' sets */
		dest = XEXP(insn, 0);
		if (REG_P(dest))
			UpdateReachingDefnInCurrBB(REGNO(dest));
		break;

	case TRAP_IF :
			/* Not yet handled */
		break;

	case ASM_INPUT :
		fprintf(stderr, "Inline ASM not yet handled...\n");
		abort();

	case UNSPEC :
	case UNSPEC_VOLATILE :
		IdentifyDependenciesIn_UNSPEC_rtx(insn);
			/* These are considered to set/use all its operands */
		break;

	case PARALLEL :
		IdentifyDependenciesIn_PARALLEL_rtx(insn);
		break;

	case SEQUENCE :
		IdentifyDependenciesIn_SEQUENCE_rtx(insn);
		break;

	case SET :
		IdentifyDependenciesIn_SET_rtx(insn);
		break;

	default :
		fprintf(stderr, "Unhandled case in \'IdentifyDependenciesInInstr\'\n");
		abort();
	}
}

void
IdentifyDependenciesInBB(BasicBlock *bb)
{
/* Make a pass over the instructions of 'bb' and identify
 * register dependencies.  Add 'sdg' edges accordingly */

	Preg     *reg;
	Iterator  iter;
	rtx       first = basic_block_head[bb->id];
	rtx       last = basic_block_end[bb->id];
	rtx 	  insn, next;
	RTX_CODE  code;

	currBB = bb;

		/* Initialise .. */
	inParallelRtx = 0;

		/* Needed to keep track of the definitions within this bb */
	defnId = bb->startDefnId;

		/* Initially, copy 'defsIn' into 'tmp' */
	reg = (Preg *)InitIter(iter, bb->activeRegs);
	while (reg != NULL) {
		Defn_CopySet(reg->defsIn, reg->tmp);
		reg = (Preg *)GetNext(iter);
	}

	currSdgNode = bb->startNode;
	for (insn = first; ; insn = next) {
		next = NEXT_INSN(insn);
		code = GET_CODE(insn);

		switch (code) {
				/* Notes, barriers and labels are useless as far
				 * register dependencies go.  They have none */
			case NOTE :
			case BARRIER :
			case CODE_LABEL :
				break;
			
				/* Now, look at all the operands in this instruction
				 * and identify dependencies */
			case CALL_INSN :
# ifdef COPYOPT_SLICING
				currSdgNode->inLdSt = 1;
				Append(&sliceNodeList, currSdgNode);
# endif
			case JUMP_INSN :
			case INSN :
				IdentifyDependenciesInInstr(PATTERN(insn));
				if (GET_CODE(PATTERN(insn)) != PARALLEL &&
					GET_CODE(PATTERN(insn)) != SEQUENCE)
				{
					assert(insn == currSdgNode->instr); /* Sanity check */
					currSdgNode = currSdgNode->next;
				}
				break;

			default :
				fprintf(stderr, "Unhandled case in \'IdentifyDependenciesInBB\'\n");
				abort();
		}

		if (insn == last)
			break;
	}
}
